from __future__ import annotations

from typing import Callable

import numpy as np
from psygnal import Signal

from whitecanvas import protocols
from whitecanvas.types import MouseEvent, Rect


class MockObject:
    def __repr__(self) -> str:
        return f"<Mock backend object at {hex(id(self))}>"


@protocols.check_protocol(protocols.CanvasProtocol)
class Canvas:
    def __init__(self):
        self._xaxis = Axis()
        self._yaxis = Axis()
        self._title = Title()
        self._xticks = Ticks()
        self._yticks = Ticks()
        self._aspect_ratio = None
        self._visible = True
        self._mouse_enabled = True
        self._obj = MockObject()

    def _plt_get_native(self):
        return self._obj

    def _plt_get_title(self):
        return self._title

    def _plt_get_xaxis(self):
        return self._xaxis

    def _plt_get_yaxis(self):
        return self._yaxis

    def _plt_get_xlabel(self):
        return self._xaxis._label

    def _plt_get_ylabel(self):
        return self._yaxis._label

    def _plt_get_xticks(self):
        return self._xticks

    def _plt_get_yticks(self):
        return self._yticks

    def _plt_reorder_layers(self, layers: list):
        pass

    def _plt_get_aspect_ratio(self) -> float | None:
        return self._aspect_ratio

    def _plt_set_aspect_ratio(self, ratio: float | None):
        self._aspect_ratio = ratio

    def _plt_add_layer(self, layer):
        pass

    def _plt_remove_layer(self, layer):
        pass

    def _plt_get_visible(self) -> bool:
        return self._visible

    def _plt_set_visible(self, visible: bool):
        self._visible = visible

    def _plt_twinx(self) -> Canvas:
        new = Canvas()

        @new._xaxis.lim_changed.connect
        def _(lims):
            with self._xaxis.lim_changed.blocked():
                self._xaxis._plt_set_limits(lims)

        @self._xaxis.lim_changed.connect
        def _(lims):
            with new._xaxis.lim_changed.blocked():
                new._xaxis._plt_set_limits(lims)

        return new

    def _plt_twiny(self) -> Canvas:
        new = Canvas()

        @new._yaxis.lim_changed.connect
        def _(lims):
            with self._yaxis.lim_changed.blocked():
                self._yaxis._plt_set_limits(lims)

        @self._yaxis.lim_changed.connect
        def _(lims):
            with new._yaxis.lim_changed.blocked():
                new._yaxis._plt_set_limits(lims)

        return new

    def _plt_inset(self, rect: Rect) -> Canvas:
        return Canvas()

    def _plt_connect_mouse_click(self, callback: Callable[[MouseEvent], None]):
        pass

    def _plt_connect_mouse_drag(self, callback: Callable[[MouseEvent], None]):
        pass

    def _plt_connect_mouse_double_click(self, callback: Callable[[MouseEvent], None]):
        pass

    def _plt_connect_mouse_release(self, callback: Callable[[MouseEvent], None]):
        pass

    def _plt_draw(self):
        pass

    def _plt_connect_xlim_changed(
        self, callback: Callable[[tuple[float, float]], None]
    ):
        self._xaxis.lim_changed.connect(callback, max_args=1)

    def _plt_connect_ylim_changed(
        self, callback: Callable[[tuple[float, float]], None]
    ):
        self._yaxis.lim_changed.connect(callback, max_args=1)

    def _plt_make_legend(self, *args, **kwargs):
        pass

    def _plt_get_mouse_enabled(self):
        return self._mouse_enabled

    def _plt_set_mouse_enabled(self, enabled: bool):
        self._mouse_enabled = enabled


@protocols.check_protocol(protocols.CanvasGridProtocol)
class CanvasGrid:
    def __init__(self, heights: list[float], widths: list[float], app: str = "default"):
        self._background_color = np.array([1, 1, 1, 1], dtype=np.float32)
        self._figsize = (100, 100)

    def _plt_add_canvas(self, row: int, col: int, rowspan: int, colspan: int) -> Canvas:
        return Canvas()

    def _plt_show(self):
        pass

    def _plt_get_background_color(self):
        return self._background_color

    def _plt_set_background_color(self, color):
        self._background_color = color

    def _plt_screenshot(self):
        raise RuntimeError("Mock backend does not support screenshots")

    def _plt_set_figsize(self, width: int, height: int):
        self._figsize = (width, height)

    def _plt_set_spacings(self, wspace: float, hspace: float):
        pass


class _SupportsText:
    def __init__(self):
        self._visible = True
        self._text = ""
        self._color = np.array([0, 0, 0, 1], dtype=np.float32)
        self._size = 10
        self._fontfamily = "Arial"

    def _plt_get_visible(self) -> bool:
        return self._visible

    def _plt_set_visible(self, visible: bool):
        self._visible = visible

    def _plt_get_text(self) -> str:
        return self._text

    def _plt_set_text(self, text: str):
        self._text = text

    def _plt_get_color(self):
        return self._color

    def _plt_set_color(self, color):
        self._color = color

    def _plt_get_size(self) -> int:
        return self._size

    def _plt_set_size(self, size: int):
        self._size = size

    def _plt_get_fontfamily(self) -> str:
        return self._fontfamily

    def _plt_set_fontfamily(self, family: str):
        self._fontfamily = family


class Title(_SupportsText):
    pass


class Label(_SupportsText):
    pass


class Axis:
    lim_changed = Signal(tuple)

    def __init__(self):
        self._visible = True
        self._limits = (0, 1)
        self._flipped = False
        self._color = np.array([0, 0, 0, 1], dtype=np.float32)
        self._label = Label()

    def _plt_get_visible(self) -> bool:
        return self._visible

    def _plt_set_visible(self, visible: bool):
        self._visible = visible

    def _plt_get_color(self):
        return self._color

    def _plt_set_color(self, color):
        self._color = color

    def _plt_get_label(self) -> Label:
        return self._label

    def _plt_flip(self) -> None:
        self._flipped = not self._flipped

    def _plt_get_limits(self) -> tuple[float, float]:
        return self._limits

    def _plt_set_limits(self, limits: tuple[float, float]):
        self._limits = limits
        self.lim_changed.emit(limits)

    def _plt_set_grid_state(self, *args, **kwargs):
        pass


class Ticks(_SupportsText):
    def __init__(self):
        super().__init__()
        self._pos = []
        self._labels = []
        self._rotation = 0.0

    def _plt_get_tick_labels(self) -> tuple[list[float], list[str]]:
        return self._pos, self._labels

    def _plt_override_labels(self, pos: list[float], labels: list[str]):
        self._pos = pos
        self._labels = labels

    def _plt_reset_override(self):
        self._pos = []
        self._labels = []

    def _plt_get_text_rotation(self) -> float:
        return self._rotation

    def _plt_set_text_rotation(self, rotation: float):
        self._rotation = rotation



from __future__ import annotations

import bokeh.models as bk_models
import numpy as np

from whitecanvas.backend.bokeh._base import HeteroLayer, SupportsMouseEvents
from whitecanvas.protocols import BarProtocol, check_protocol


@check_protocol(BarProtocol)
class Bars(HeteroLayer[bk_models.Quad], SupportsMouseEvents):
    def __init__(self, xlow, xhigh, ylow, yhigh):
        ndata = len(xlow)
        self._visible = True
        self._data = bk_models.ColumnDataSource(
            data={
                "x0": xlow,
                "x1": xhigh,
                "y0": ylow,
                "y1": yhigh,
                "face_color": ["blue"] * ndata,
                "edge_color": ["black"] * ndata,
                "width": np.zeros(ndata),
                "pattern": [" "] * ndata,
                "style": ["solid"] * ndata,
                "hovertexts": np.array([""] * len(xlow)),
            }
        )
        self._model = bk_models.Quad(
            left="x0",
            right="x1",
            bottom="y0",
            top="y1",
            fill_alpha=1.0,
            line_color="edge_color",
            line_width="width",
            fill_color="face_color",
            hatch_pattern="pattern",
            line_dash="style",
        )

    ##### XXYYDataProtocol #####
    def _plt_get_data(self):
        return (
            self._data.data["x0"],
            self._data.data["x1"],
            self._data.data["y0"],
            self._data.data["y1"],
        )

    def _plt_set_data(self, x0, x1, y0, y1):
        cur_data = self._data.data.copy()
        ndata: int = cur_data["x0"].size
        cur_data.update({"x0": x0, "x1": x1, "y0": y0, "y1": y1})
        cols_to_update = [
            "face_color", "edge_color", "width", "pattern", "style", "hovertexts"
        ]  # fmt: skip
        if x0.size < ndata:
            for key in cols_to_update:
                cur_data[key] = cur_data[key][: x0.size]
        elif x0.size > ndata:
            for key in cols_to_update:
                cur_data[key] = np.concatenate(
                    [cur_data[key], np.full(x0.size - ndata, cur_data[key][-1])]
                )
        self._data.data = cur_data

    def _plt_get_ndata(self) -> int:
        return len(self._data.data["x0"])



from __future__ import annotations

from typing import Generic, TypeVar

from whitecanvas.layers._primitive import Band, Markers
from whitecanvas.layers._primitive.line import _SingleLine
from whitecanvas.layers.group._collections import LayerContainer
from whitecanvas.types import ColorType, Hatch, Symbol, XYData, _Void

_void = _Void()
_L = TypeVar("_L", bound=_SingleLine)


class LineBand(LayerContainer, Generic[_L]):
    """
    Group of Line, Band and Markers.

    Properties:
        line: The central line layer
        band: The band region around the central line
        markers: The markers at the data points
    """

    def __init__(
        self,
        line: _L,
        band: Band,
        markers: Markers | None = None,
        name: str | None = None,
    ):
        if markers is None:
            markers = Markers([], [], name="markers")
        super().__init__([line, band, markers], name=name)

    @property
    def line(self) -> _L:
        """The central line layer."""
        return self._children[0]

    @property
    def band(self) -> Band:
        """The band region layer."""
        return self._children[1]

    @property
    def markers(self) -> Markers:
        """The markers layer."""
        return self._children[2]

    @property
    def data(self) -> XYData:
        """Current data of the central line."""
        return self.line.data

    def with_markers(
        self,
        symbol: Symbol | str = Symbol.CIRCLE,
        size: float = 10,
        color: ColorType | _Void = _void,
        alpha: float = 1.0,
        hatch: str | Hatch = Hatch.SOLID,
    ) -> LineBand:
        """Add markers at the data points."""
        if color is _void:
            color = self.line.color
        self.markers.update(
            symbol=symbol, size=size, color=color, alpha=alpha, hatch=hatch
        )
        return self



from __future__ import annotations

import sys
import warnings
import weakref
from typing import TYPE_CHECKING, Callable

import numpy as np
from plotly import graph_objects as go

from whitecanvas.backend.plotly._labels import Axis3D, AxisLabel3D, Ticks3D
from whitecanvas.backend.plotly.canvas import Canvas


class Canvas3D(Canvas):
    def __init__(
        self,
        fig: go.Figure,
        *,
        row: int = 0,
        col: int = 0,
        secondary_y: bool = False,
        app: str = "default",
    ):
        super().__init__(fig, row=row, col=col, secondary_y=secondary_y, app=app)
        self._xaxis = Axis3D(self, axis="xaxis")
        self._yaxis = Axis3D(self, axis="yaxis")
        self._zaxis = Axis3D(self, axis="zaxis")
        self._xticks = Ticks3D(self, axis="xaxis")
        self._yticks = Ticks3D(self, axis="yaxis")
        self._zticks = Ticks3D(self, axis="zaxis")
        self._xlabel = AxisLabel3D(self, axis="xaxis")
        self._ylabel = AxisLabel3D(self, axis="yaxis")
        self._zlabel = AxisLabel3D(self, axis="zaxis")
        self._is_aspect_locked = False

    def _plt_get_zaxis(self):
        return self._zaxis

    def _plt_get_zlabel(self):
        return self._zlabel

    def _plt_get_zticks(self):
        return self._zticks

    def _plt_get_aspect_locked(self) -> bool:
        return self._is_aspect_locked

    def _plt_set_aspect_locked(self, locked: bool):
        if locked:
            self._fig.update_layout(scene_aspectmode="auto")
        else:
            self._fig.update_layout(scene_aspectmode="data")

    def _plt_connect_xlim_changed(self, callback):
        propname = f"scene.{self._xaxis.name}.range"
        self._fig.layout.on_change(lambda _, lim: callback(lim), propname)

    def _plt_connect_ylim_changed(self, callback):
        propname = f"scene.{self._yaxis.name}.range"
        self._fig.layout.on_change(lambda _, lim: callback(lim), propname)

    def _plt_connect_zlim_changed(self, callback):
        propname = f"scene.{self._zaxis.name}.range"
        self._fig.layout.on_change(lambda _, lim: callback(lim), propname)



from __future__ import annotations

from dataclasses import dataclass

import numpy as np
from cmap import Color

from whitecanvas.types import Hatch, LineStyle, Symbol


@dataclass
class LegendItem:
    pass


@dataclass
class FaceInfo:
    color: Color
    hatch: Hatch = Hatch.SOLID

    def __post_init__(self):
        self.color = np.array(Color(self.color).rgba)
        self.hatch = Hatch(self.hatch)


@dataclass
class EdgeInfo:
    color: Color
    width: float
    style: LineStyle = LineStyle.SOLID

    def __post_init__(self):
        self.color = np.array(Color(self.color).rgba)
        self.style = LineStyle(self.style)


class EmptyLegendItem(LegendItem):
    pass


class LineLegendItem(EdgeInfo, LegendItem):
    pass


@dataclass
class ErrorLegendItem(EdgeInfo):
    color: Color
    width: float
    style: LineStyle = LineStyle.SOLID
    capsize: float = 0.0


@dataclass
class MarkersLegendItem(LegendItem):
    symbol: Symbol
    size: float
    face: FaceInfo
    edge: EdgeInfo


@dataclass
class BarLegendItem(LegendItem):
    face: FaceInfo
    edge: EdgeInfo


@dataclass
class PlotLegendItem(LegendItem):
    line: LineLegendItem
    markers: MarkersLegendItem

    def __post_init__(self):
        if not isinstance(self.line, LineLegendItem):
            raise ValueError(f"line got {type(self.line)}")
        if not isinstance(self.markers, MarkersLegendItem):
            raise ValueError(f"markers got {type(self.markers)}")


@dataclass
class LineErrorLegendItem(LegendItem):
    line: LineLegendItem
    xerr: ErrorLegendItem | None = None
    yerr: ErrorLegendItem | None = None

    def __post_init__(self):
        if not isinstance(self.line, LineLegendItem):
            raise ValueError(f"line got {type(self.line)}")
        if self.xerr is not None and not isinstance(self.xerr, ErrorLegendItem):
            raise ValueError(f"xerr got {type(self.xerr)}")
        if self.yerr is not None and not isinstance(self.yerr, ErrorLegendItem):
            raise ValueError(f"yerr got {type(self.yerr)}")


@dataclass
class MarkerErrorLegendItem(LegendItem):
    markers: MarkersLegendItem
    xerr: ErrorLegendItem | None = None
    yerr: ErrorLegendItem | None = None

    def __post_init__(self):
        if not isinstance(self.markers, MarkersLegendItem):
            raise ValueError(f"markers got {type(self.markers)}")
        if self.xerr is not None and not isinstance(self.xerr, ErrorLegendItem):
            raise ValueError(f"xerr got {type(self.xerr)}")
        if self.yerr is not None and not isinstance(self.yerr, ErrorLegendItem):
            raise ValueError(f"yerr got {type(self.yerr)}")


@dataclass
class PlotErrorLegendItem(LegendItem):
    plot: PlotLegendItem
    xerr: ErrorLegendItem | None = None
    yerr: ErrorLegendItem | None = None

    def __post_init__(self):
        if not isinstance(self.plot, PlotLegendItem):
            raise ValueError(f"plot got {type(self.plot)}")
        if self.xerr is not None and not isinstance(self.xerr, ErrorLegendItem):
            raise ValueError(f"xerr got {type(self.xerr)}")
        if self.yerr is not None and not isinstance(self.yerr, ErrorLegendItem):
            raise ValueError(f"yerr got {type(self.yerr)}")


@dataclass
class StemLegendItem(LegendItem):
    line: LineLegendItem
    markers: MarkersLegendItem

    def __post_init__(self):
        if not isinstance(self.line, LineLegendItem):
            raise ValueError(f"line got {type(self.line)}")
        if not isinstance(self.markers, MarkersLegendItem):
            raise ValueError(f"markers got {type(self.markers)}")


@dataclass
class LegendItemCollection(LegendItem):
    items: list[tuple[str, LegendItem]]


@dataclass
class TitleItem(LegendItem):
    pass



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Iterable, Iterator

import numpy as np
from numpy.typing import ArrayLike, NDArray
from psygnal import Signal

from whitecanvas.backend import Backend
from whitecanvas.layers import _legend
from whitecanvas.layers._mixin import (
    EnumArray,
    MultiEdge,
    MultiFace,
    SinglePropertyEdgeBase,
    SinglePropertyFaceBase,
)
from whitecanvas.layers._primitive import Markers
from whitecanvas.layers.group._collections import (
    LayerCollectionBase,
    LayerContainerEvents,
)
from whitecanvas.theme import get_theme
from whitecanvas.types import (
    ArrayLike1D,
    ColorType,
    Hatch,
    LineStyle,
    Symbol,
    XYData,
    _Void,
)
from whitecanvas.utils.normalize import (
    as_any_1d_array,
    as_array_1d,
    as_color_array,
    parse_texts,
)
from whitecanvas.utils.type_check import is_real_number

if TYPE_CHECKING:
    from typing_extensions import Self

_void = _Void()
_Markers = Markers[MultiFace, MultiEdge, NDArray[np.floating]]


class MarkerCollectionEvents(LayerContainerEvents):
    size = Signal(object)
    symbol = Signal(object)


class MarkerCollectionFace(SinglePropertyFaceBase):
    _layer: MarkerCollection

    def _iter_markers(self) -> Iterator[tuple[NDArray[np.bool_], _Markers]]:
        for i, layer in enumerate(self._layer):
            yield self._layer._slices[i], layer

    @property
    def color(self) -> NDArray[np.floating]:
        """Face color of the markers."""
        out = np.empty((self._layer.ndata, 4), dtype=np.float32)
        for sl, layer in self._iter_markers():
            out[sl] = layer.face.color
        return out

    @color.setter
    def color(self, color):
        colors = as_color_array(color, self._layer.ndata)
        for sl, layer in self._iter_markers():
            layer.face.color = colors[sl]
        self.events.color.emit(colors)

    @property
    def hatch(self) -> EnumArray[Hatch]:
        """Face fill hatch."""
        out = np.empty(self._layer.ndata, dtype=object)
        for sl, layer in self._iter_markers():
            out[sl] = layer.face.hatch
        return out

    @hatch.setter
    def hatch(self, hatch: str | Hatch | Iterable[str | Hatch]):
        if isinstance(hatch, (str, Hatch)):
            h = Hatch(hatch)
            for layer in self._layer:
                layer.face.hatch = h
            self.events.hatch.emit(h)
        else:
            _hatches = np.array([Hatch(h) for h in hatch], dtype=object)
            markers = list(self._layer)
            for i, layer in enumerate(markers):
                sl = self._layer._slices[i]
                _hat = _hatches[sl]
                _hat_unique = list(set(_hat))
                if len(_hat_unique) == 1:
                    layer.face.hatch = _hat_unique[0]
                else:
                    new_mks = self._layer._split_markers(i, _hat, _hat_unique)
                    for mk, each in zip(new_mks, _hat_unique):
                        mk.face.hatch = each
            self.events.hatch.emit(_hatches)

    @property
    def alpha(self) -> float:
        return self.color[:, 3]

    @alpha.setter
    def alpha(self, value):
        color = self.color.copy()
        color[:, 3] = value
        self.color = color

    def update(
        self,
        *,
        color: ColorType | _Void = _void,
        hatch: Hatch | str | _Void = _void,
        alpha: float | _Void = _void,
    ) -> MarkerCollection:
        """
        Update the face properties.

        Parameters
        ----------
        color : ColorType, optional
            Color of the face.
        hatch : FacePattern, optional
            Fill hatch of the face.
        alpha : float, optional
            Alpha value of the face.
        """
        if color is not _void:
            self.color = color
        if hatch is not _void:
            self.hatch = hatch
        if alpha is not _void:
            self.alpha = alpha
        return self._layer

    def _as_legend_info(self):
        if self.color.size == 0:
            return _legend.EmptyLegendItem()
        color = self.color[0]
        hatch = self.hatch[0]
        return _legend.FaceInfo(color, hatch)


class MarkerCollectionEdge(SinglePropertyEdgeBase):
    _layer: MarkerCollection

    def _iter_markers(self) -> Iterator[tuple[NDArray[np.bool_], _Markers]]:
        for i, layer in enumerate(self._layer):
            yield self._layer._slices[i], layer

    @property
    def color(self) -> NDArray[np.floating]:
        """Edge color of the markers."""
        out = np.empty((self._layer.ndata, 4), dtype=np.float32)
        for sl, layer in self._iter_markers():
            out[sl] = layer.edge.color
        return out

    @color.setter
    def color(self, color):
        colors = as_color_array(color, self._layer.ndata)
        for sl, layer in self._iter_markers():
            layer.edge.color = colors[sl]
        self.events.color.emit(colors)

    @property
    def width(self) -> NDArray[np.float32]:
        """Edge widths."""
        out = np.empty(self._layer.ndata, dtype=np.float32)
        for sl, layer in self._iter_markers():
            out[sl] = layer.edge.width
        return out

    @width.setter
    def width(self, width: float | Iterable[float]):
        if is_real_number(width):
            for _, layer in self._iter_markers():
                layer.edge.width = width
            self.events.width.emit(width)
        else:
            widths = as_array_1d(width)
            for sl, layer in self._iter_markers():
                layer.edge.width = widths[sl]
            self.events.width.emit(widths)

    @property
    def style(self) -> EnumArray[LineStyle]:
        """Edge line style."""
        out = np.empty(self._layer.ndata, dtype=object)
        for sl, layer in self._iter_markers():
            out[sl] = layer.edge.style
        return out

    @style.setter
    def style(self, style: str | Hatch | Iterable[str | Hatch]):
        styles = as_any_1d_array(style, self._layer.ndata, dtype=object)
        for sl, layer in self._iter_markers():
            layer.edge.style = styles[sl]
        self.events.style.emit(styles)

    @property
    def alpha(self) -> float:
        return self.color[:, 3]

    @alpha.setter
    def alpha(self, value):
        color = self.color.copy()
        color[:, 3] = value
        self.color = color

    def update(
        self,
        *,
        color: ColorType | _Void = _void,
        style: LineStyle | str | _Void = _void,
        width: float | _Void = _void,
        alpha: float | _Void = _void,
    ):
        if color is not _void:
            self.color = color
        if style is not _void:
            self.style = style
        if width is not _void:
            self.width = width
        if alpha is not _void:
            self.alpha = alpha
        return self._layer

    def _as_legend_info(self):
        if self.color.size == 0:
            return _legend.EmptyLegendItem()
        color = self.color[0]
        width = self.width[0]
        style = self.style[0]
        return _legend.EdgeInfo(color, width, style)


class MarkerCollection(LayerCollectionBase[_Markers]):
    events: MarkerCollectionEvents
    _events_class = MarkerCollectionEvents

    def __init__(
        self,
        layers: list[_Markers],
        name: str | None = None,
    ):
        super().__init__(layers, name)
        self._face_namespace = MarkerCollectionFace(self)
        self._edge_namespace = MarkerCollectionEdge(self)
        sizes = [layer.ndata for layer in layers]
        cum_size = np.cumsum([0, *sizes])
        ndata = cum_size[-1]
        self._slices = [np.zeros(ndata, dtype=np.bool_) for _ in sizes]
        for i, sl in enumerate(self._slices):
            sl[cum_size[i] : cum_size[i + 1]] = True

    @classmethod
    def from_arrays(
        cls,
        xdata: ArrayLike1D,
        ydata: ArrayLike1D,
        name: str | None = None,
        backend: str | Backend | None = None,
    ) -> Self:
        markers = Markers(xdata, ydata, backend=backend)._as_all_multi()
        return cls([markers], name=name)

    @property
    def face(self) -> MarkerCollectionFace:
        return self._face_namespace

    @property
    def edge(self) -> MarkerCollectionEdge:
        return self._edge_namespace

    @property
    def ndata(self) -> int:
        """Number of data points."""
        return sum(layer.ndata for layer in self)

    @property
    def data(self) -> XYData:
        """All the data of the markers."""
        ndata = self.ndata
        xs = np.empty(ndata, dtype=np.float32)
        ys = np.empty(ndata, dtype=np.float32)
        for i, layer in enumerate(self):
            data = layer.data
            sl = self._slices[i]
            xs[sl] = data.x
            ys[sl] = data.y
        return XYData(xs, ys)

    @data.setter
    def data(self, data: XYData):
        data = self._norm_layer_data(data)
        for i, layer in enumerate(self):
            sl = self._slices[i]
            x0 = data.x[sl]
            y0 = data.y[sl]
            layer.data = XYData(x0, y0)

    def _norm_layer_data(self, data: Any) -> XYData:
        if isinstance(data, np.ndarray):
            if data.ndim != 2 or data.shape[1] != 2:
                raise ValueError("Expected a (N, 2) array")
            xdata, ydata = data[:, 0], data[:, 1]
        else:
            xdata, ydata = data
        x0, y0 = self.data
        if xdata is not None:
            x0 = as_array_1d(xdata)
        if ydata is not None:
            y0 = as_array_1d(ydata)
        if x0.size != y0.size:
            raise ValueError(
                "Expected xdata and ydata to have the same size, "
                f"got {x0.size} and {y0.size}"
            )
        return XYData(x0, y0)

    def set_data(
        self,
        xdata: ArrayLike | None = None,
        ydata: ArrayLike | None = None,
    ):
        if xdata is None:
            xdata = self.data.x
        if ydata is None:
            ydata = self.data.y
        self.data = XYData(xdata, ydata)

    @property
    def symbol(self) -> EnumArray[Symbol]:
        """Symbol of the markers."""
        out = np.empty(self.ndata, dtype=object)
        for i, layer in enumerate(self):
            sl = self._slices[i]
            out[sl] = layer.symbol
        return out

    @symbol.setter
    def symbol(self, symbol: str | Symbol | Iterable[str | Symbol]):
        if isinstance(symbol, (str, Symbol)):
            sym = Symbol(symbol)
            for layer in self:
                layer.symbol = sym
            self.events.symbol.emit(sym)
        else:
            _symbols = np.array([Symbol(s) for s in symbol], dtype=object)
            markers = list(self)
            for i, layer in enumerate(markers):
                sl = self._slices[i]
                _sym = _symbols[sl]
                _sym_unique = list(set(_sym))
                if len(_sym_unique) == 1:
                    layer.symbol = _sym_unique[0]
                else:
                    new_mks = self._split_markers(i, _sym, _sym_unique)
                    for mk, each in zip(new_mks, _sym_unique):
                        mk.symbol = each
            self.events.symbol.emit(_symbols)

    @property
    def size(self) -> NDArray[np.float32]:
        out = np.empty(self.ndata, dtype=np.float32)
        for i, layer in enumerate(self):
            sl = self._slices[i]
            out[sl] = layer.size
        return out

    @size.setter
    def size(self, size: float | Iterable[float]):
        if is_real_number(size):
            for i, layer in enumerate(self):
                sl = self._slices[i]
                layer.size = size
            self.events.size.emit(size)
        else:
            size = as_array_1d(size)
            for i, layer in enumerate(self):
                sl = self._slices[i]
                layer.size = size[sl]
            self.events.size.emit(size)

    def with_face(
        self,
        *,
        color: ColorType | _Void = _void,
        hatch: Hatch | str = Hatch.SOLID,
        alpha: float = 1,
    ) -> Self:
        """Update the face properties."""
        self.face.update(color=color, hatch=hatch, alpha=alpha)
        return self

    def with_edge(
        self,
        *,
        color: ColorType | None = None,
        width: float = 1.0,
        style: LineStyle | str = LineStyle.SOLID,
        alpha: float = 1,
    ) -> Self:
        """Update the edge properties."""
        if color is None:
            color = get_theme().foreground_color
        self.edge.update(color=color, style=style, width=width, alpha=alpha)
        return self

    def with_hover_text(self, text: str | Iterable[Any]) -> Self:
        """Add hover text to the markers."""
        if isinstance(text, str):
            texts = np.full(self.ndata, text, dtype=object)
        else:
            texts = np.array([str(t) for t in text], dtype=object)
        if len(texts) != self.ndata:
            raise ValueError(
                "Expected text to have the same size as the data, "
                f"got {len(texts)} and {self.ndata}"
            )
        return self._set_hover_text_unsafe(texts)

    def with_hover_template(self, template: str, extra: Any | None = None) -> Self:
        """Add hover template to the markers."""
        xs, ys = self.data
        if xs.size == 0:  # empty layer
            return self
        if self._backend_name in ("plotly", "bokeh"):  # conversion for HTML
            template = template.replace("\n", "<br>")
        params = parse_texts(template, xs.size, extra)
        # set default format keys
        params.setdefault("x", xs)
        params.setdefault("y", ys)
        if "i" not in params:
            params["i"] = np.arange(xs.size)
        texts = [
            template.format(**{k: v[i] for k, v in params.items()})
            for i in range(xs.size)
        ]
        texts = np.array(texts, dtype=object)
        return self._set_hover_text_unsafe(texts)

    def _set_hover_text_unsafe(self, texts: NDArray[np.object_]):
        for i, layer in enumerate(self):
            sl = self._slices[i]
            layer._backend._plt_set_hover_text(texts[sl])
        return self

    def _split_markers(
        self,
        idx: int,
        spec: NDArray[Any],
        spec_unique: NDArray[Any] | None = None,
    ) -> list[_Markers]:
        """Split the marker at index i."""
        layer = self[idx]
        all_data = layer.data
        if spec_unique is None:
            spec_unique = set(spec)
        new_markers: list[_Markers] = []
        new_slices = self._slices.copy()
        new_slices.pop(idx)
        for each in spec_unique:
            mask = spec == each
            markers = Markers(
                all_data.x[mask],
                all_data.y[mask],
                symbol=layer.symbol,
                backend=self._backend_name,
            )
            markers.with_size_multi(
                layer.size[mask],
            ).with_face_multi(
                color=layer.face.color[mask],
                hatch=layer.face.hatch[mask],
            ).with_edge_multi(
                color=layer.edge.color[mask],
                width=layer.edge.width[mask],
                style=layer.edge.style[mask],
            )
            new_markers.append(markers)
            new_slices.append(mask)

        self.pop(idx)
        self.extend(new_markers)
        self._slices = new_slices
        return new_markers

    def _as_legend_item(self):
        if self._children:
            return self._children[0]._as_legend_item()
        return _legend.EmptyLegendItem()



from __future__ import annotations

from typing import Iterable

import numpy as np
from numpy.typing import NDArray

from whitecanvas.backend import Backend
from whitecanvas.layers._mixin import (
    AbstractFaceEdgeMixin,
    EnumArray,
    MultiEdge,
    MultiFace,
    MultiPropertyEdgeBase,
    MultiPropertyFaceBase,
)
from whitecanvas.layers._primitive import Bars, MultiLine
from whitecanvas.layers.group._cat_utils import check_array_input
from whitecanvas.layers.group._collections import LayerContainer, RichContainerEvents
from whitecanvas.theme import get_theme
from whitecanvas.types import (
    ArrayLike1D,
    ColorType,
    Hatch,
    LineStyle,
    Orientation,
    OrientationLike,
    _Void,
)
from whitecanvas.utils.normalize import as_any_1d_array, as_color_array

_void = _Void()


class BoxPlot(LayerContainer, AbstractFaceEdgeMixin["BoxFace", "BoxEdge"]):
    """
    A group for boxplot.

    Children layers are:
    - Bars (boxes)
    - MultiLine (whiskers)
    - MultiLine (median line)

     ──┬──  <-- max
       │
    ┌──┴──┐ <-- 75% quantile
    │  o  │ <-- mean
    ╞═════╡ <-- median
    └──┬──┘ <-- 25% quantile
       │
     ──┴──  <-- min
    """

    events: RichContainerEvents
    _events_class = RichContainerEvents

    def __init__(
        self,
        boxes: Bars[MultiFace, MultiEdge],
        whiskers: MultiLine,
        medians: MultiLine,
        *,
        name: str | None = None,
        orient: OrientationLike = "vertical",
        capsize: float = 0.15,
    ):
        super().__init__([boxes, whiskers, medians], name=name)
        AbstractFaceEdgeMixin.__init__(self, BoxFace(self), BoxEdge(self))
        self._orient = Orientation.parse(orient)
        self._capsize = capsize
        self._init_events()

    @property
    def boxes(self) -> Bars[MultiFace, MultiEdge]:
        """The boxes layer (Bars)."""
        return self._children[0]

    @property
    def whiskers(self) -> MultiLine:
        """The whiskers layer (MultiLine)."""
        return self._children[1]

    @property
    def medians(self) -> MultiLine:
        """The median line layer (MultiLine)."""
        return self._children[2]

    @classmethod
    def from_arrays(
        cls,
        x: list[float],
        data: list[ArrayLike1D],
        *,
        name: str | None = None,
        orient: OrientationLike = "vertical",
        extent: float = 0.3,
        capsize: float = 0.15,
        color: ColorType | list[ColorType] = "blue",
        alpha: float = 1.0,
        hatch: str | Hatch = Hatch.SOLID,
        backend: str | Backend | None = None,
    ):
        x, data = check_array_input(x, data)
        ori = Orientation.parse(orient)
        color = as_color_array(color, len(x))
        agg_values: list[NDArray[np.number]] = []
        for d in data:
            agg_values.append(np.quantile(d, [0, 0.25, 0.5, 0.75, 1]))
        agg_arr = np.stack(agg_values, axis=1)
        box = Bars(
            x, agg_arr[3] - agg_arr[1], agg_arr[1], name=name, orient=ori,
            extent=extent, backend=backend,
        ).with_face_multi(
            hatch=hatch, color=color, alpha=alpha,
        ).with_edge_multi(color="black")  # fmt: skip
        if ori.is_vertical:
            segs = _xyy_to_segments(
                x, agg_arr[0], agg_arr[1], agg_arr[3], agg_arr[4], capsize
            )
            medsegs = [
                [(x0 - extent / 2, y0), (x0 + extent / 2, y0)]
                for x0, y0 in zip(x, agg_arr[2])
            ]
        else:
            segs = _yxx_to_segments(
                x, agg_arr[0], agg_arr[1], agg_arr[3], agg_arr[4], capsize
            )
            medsegs = [
                [(x0, y0 - extent / 2), (x0, y0 + extent / 2)]
                for x0, y0 in zip(x, agg_arr[2])
            ]
        whiskers = MultiLine(
            segs, name=name, style=LineStyle.SOLID, alpha=alpha, backend=backend,
            color="black",
        )  # fmt: skip
        medians = MultiLine(
            medsegs, name="medians", color="black", alpha=alpha, backend=backend,
        )  # fmt: skip

        return cls(box, whiskers, medians, name=name, orient=ori, capsize=capsize)

    @property
    def orient(self) -> Orientation:
        """Orientation of the boxplot."""
        return self._orient

    def move(self, shift: float, autoscale: bool = True) -> BoxPlot:
        """Move the layer by the given shift."""
        self.boxes.set_data(xdata=self.boxes.data.x + shift)
        if self.orient.is_vertical:
            _wdata = []
            for seg in self.whiskers.data:
                _wdata.append(np.stack([seg[:, 0] + shift, seg[:, 1]], axis=0))
            self.whiskers.data = _wdata
            _mdata = []
            for seg in self.medians.data:
                _mdata.append(np.stack([seg[:, 0] + shift, seg[:, 1]], axis=0))
            self.medians.data = _mdata
        else:
            _wdata = []
            for seg in self.whiskers.data:
                _wdata.append(np.stack([seg[:, 0], seg[:, 1] + shift], axis=0))
            self.whiskers.data = _wdata
            _mdata = []
            for seg in self.medians.data:
                _mdata.append(np.stack([seg[:, 0], seg[:, 1] + shift], axis=0))
            self.medians.data = _mdata
        if autoscale and (canvas := self._canvas_ref()):
            canvas._autoscale_for_layer(self, pad_rel=0.025)
        return self

    def _update_data(self, agg_arr: NDArray[np.number]):
        x = self.boxes.data.x
        extent = self.boxes.bar_width
        self.boxes.set_data(ydata=agg_arr[3] - agg_arr[1], bottom=agg_arr[1])
        if self.orient.is_vertical:
            segs = _xyy_to_segments(
                x, agg_arr[0], agg_arr[1], agg_arr[3], agg_arr[4], self._capsize
            )
            medsegs = [
                np.array([(x0 - extent / 2, y0), (x0 + extent / 2, y0)])
                for x0, y0 in zip(x, agg_arr[2])
            ]
        else:
            segs = _yxx_to_segments(
                x, agg_arr[0], agg_arr[1], agg_arr[3], agg_arr[4], self._capsize
            )
            medsegs = [
                np.array([(x0, y0 - extent / 2), (x0, y0 + extent / 2)])
                for x0, y0 in zip(x, agg_arr[2])
            ]
        self.whiskers.data = segs
        self.medians.data = medsegs
        return None

    def _get_sep_values(self) -> NDArray[np.number]:
        """(5, N) array of min, 25%, 50%, 75%, max."""
        idx = 1 if self.orient.is_vertical else 0
        stop = self.boxes.ndata
        _min = [seg[0, idx] for seg in self.whiskers.data[:stop]]
        _p25 = self.boxes.bottom
        _median = [seg[0, idx] for seg in self.medians.data]
        _p75 = self.boxes.top
        _max = [seg[1, idx] for seg in self.whiskers.data[stop : stop * 2]]
        return np.stack([_min, _p25, _median, _p75, _max], axis=0)

    def _make_sure_hatch_visible(self):
        _is_no_width = self.edge.width == 0
        if np.any(_is_no_width):
            ec = get_theme().foreground_color
            self.edge.width = np.where(_is_no_width, 1, self.edge.width)
            self.edge.color = np.where(_is_no_width, ec, self.edge.color)

    def _xndata(self) -> int:
        nboxes = self.boxes.ndata
        nlines = self.whiskers.ndata
        assert nboxes * 2 == nlines or nboxes * 4 == nlines, f"{nboxes=}, {nlines=}"
        return nlines // nboxes


def _xyy_to_segments(
    x: ArrayLike1D,
    y0: ArrayLike1D,
    y1: ArrayLike1D,
    y2: ArrayLike1D,
    y3: ArrayLike1D,
    capsize: float,
):
    """
    ──┬──  <-- y3
      │    <-- y2

      │    <-- y1
    ──┴──  <-- y0
      ↑
      x
    """
    v0 = np.stack([x, y0], axis=1)
    v1 = np.stack([x, y1], axis=1)
    v2 = np.stack([x, y2], axis=1)
    v3 = np.stack([x, y3], axis=1)
    segments_0 = [np.stack([s0, s1]) for s0, s1 in zip(v0, v1)]
    segments_1 = [np.stack([s2, s3]) for s2, s3 in zip(v2, v3)]
    if capsize > 0:
        _c = np.array([capsize / 2, 0])
        cap0 = [np.stack([s0 - _c, s0 + _c]) for s0 in v0]
        cap1 = [np.stack([s3 - _c, s3 + _c]) for s3 in v3]
    else:
        cap0 = []
        cap1 = []
    return segments_0 + segments_1 + cap0 + cap1


def _yxx_to_segments(
    y: ArrayLike1D,
    x0: ArrayLike1D,
    x1: ArrayLike1D,
    x2: ArrayLike1D,
    x3: ArrayLike1D,
    capsize: float,
):
    """
    |        |
     ───  ───  <-- y
    |        |
    ↑  ↑  ↑  ↑
    x0 x1 x2 x3
    """
    v0 = np.stack([x0, y], axis=1)
    v1 = np.stack([x1, y], axis=1)
    v2 = np.stack([x2, y], axis=1)
    v3 = np.stack([x3, y], axis=1)
    segments_0 = [np.stack([s0, s1]) for s0, s1 in zip(v0, v1)]
    segments_1 = [np.stack([s2, s3]) for s2, s3 in zip(v2, v3)]

    if capsize > 0:
        _c = np.array([0, capsize / 2])
        cap0 = [np.stack([s0 - _c, s0 + _c]) for s0 in v0]
        cap1 = [np.stack([s3 - _c, s3 + _c]) for s3 in v3]
    else:
        cap0 = []
        cap1 = []
    return segments_0 + segments_1 + cap0 + cap1


class BoxFace(MultiPropertyFaceBase):
    _layer: BoxPlot

    @property
    def color(self) -> NDArray[np.floating]:
        """Face color of the bar."""
        return self._layer.boxes.face.color

    @color.setter
    def color(self, color):
        ndata = self._layer.boxes.ndata
        col = as_color_array(color, ndata)
        self._layer.boxes.face.color = col
        self.events.color.emit(col)

    @property
    def hatch(self) -> EnumArray[Hatch]:
        """Face fill hatch."""
        return self._layer.boxes.face.hatch

    @hatch.setter
    def hatch(self, hatch: str | Hatch | Iterable[str | Hatch]):
        ndata = self._layer.boxes.ndata
        hatches = as_any_1d_array(hatch, ndata, dtype=object)
        self._layer.boxes.face.hatch = hatches
        self.events.hatch.emit(hatches)


class BoxEdge(MultiPropertyEdgeBase):
    _layer: BoxPlot

    def _xndata(self) -> int:
        return self._layer._xndata()

    @property
    def color(self) -> NDArray[np.floating]:
        """Edge color of the box plot."""
        return self._layer.boxes.edge.color

    @color.setter
    def color(self, color: ColorType):
        ndata = self._layer.boxes.ndata
        col = as_color_array(color, ndata)
        self._layer.boxes.edge.color = col
        self._layer.whiskers.color = np.concatenate([col] * self._xndata(), axis=0)
        self._layer.medians.color = col
        self.events.color.emit(col)

    @property
    def width(self) -> NDArray[np.float32]:
        """Edge widths."""
        return self._layer.boxes.edge.width

    @width.setter
    def width(self, width: float):
        ndata = self._layer.boxes.ndata
        widths = as_any_1d_array(width, ndata, dtype=np.float32)
        self._layer.boxes.edge.width = widths
        self._layer.whiskers.width = np.tile(widths, self._xndata())
        self._layer.medians.width = widths
        self.events.width.emit(widths)

    @property
    def style(self) -> EnumArray[LineStyle]:
        """Edge styles."""
        return self._layer.boxes.edge.style

    @style.setter
    def style(self, style: str | LineStyle):
        ndata = self._layer.boxes.ndata
        if isinstance(style, (str, LineStyle)):
            styles = np.full(ndata, LineStyle(style), dtype=object)
        else:
            styles = np.array(style, dtype=object)
            if styles.shape != (ndata,):
                raise ValueError("Invalid shape of the style array.")
        self._layer.boxes.edge.style = styles
        self._layer.whiskers.style = np.tile(styles, self._xndata())
        self._layer.medians.style = styles
        self.events.style.emit(style)



from __future__ import annotations

import warnings

import numpy as np
from numpy.typing import NDArray
from plotly import graph_objects as go

from whitecanvas.backend.plotly._base import (
    PlotlyHoverableLayer,
    from_plotly_linestyle,
    to_plotly_linestyle,
)
from whitecanvas.types import LineStyle
from whitecanvas.utils.normalize import arr_color, rgba_str_color


class MonoLine3D(PlotlyHoverableLayer[go.Scatter3d]):
    def __init__(self, xdata, ydata, zdata):
        ndata = len(xdata)
        self._props = {
            "x": xdata,
            "y": ydata,
            "z": zdata,
            "mode": "lines",
            "line": {"color": "blue", "width": 1, "dash": "solid"},
            "type": "scatter3d",
            "showlegend": False,
            "visible": True,
            "customdata": [""] * ndata,
            "hovertemplate": "%{customdata}<extra></extra>",
        }
        PlotlyHoverableLayer.__init__(self)

    def _plt_get_data(self):
        return self._props["x"], self._props["y"]

    def _plt_set_data(self, xdata, ydata):
        self._props["x"] = xdata
        self._props["y"] = ydata

    def _plt_get_ndata(self) -> int:
        return len(self._props["x"])

    def _plt_get_edge_width(self) -> float:
        return self._props["line"]["width"]

    def _plt_set_edge_width(self, width: float):
        self._props["line"]["width"] = width

    def _plt_get_edge_style(self) -> LineStyle:
        return from_plotly_linestyle(self._props["line"]["dash"])

    def _plt_set_edge_style(self, style: LineStyle):
        self._props["line"]["dash"] = to_plotly_linestyle(style)

    def _plt_get_edge_color(self) -> NDArray[np.float32]:
        return arr_color(self._props["line"]["color"])

    def _plt_set_edge_color(self, color: NDArray[np.float32]):
        self._props["line"]["color"] = rgba_str_color(color)

    def _plt_get_antialias(self) -> bool:
        return False

    def _plt_set_antialias(self, antialias: bool):
        pass



from __future__ import annotations

from typing import Any, Iterable, Sequence

import numpy as np
from numpy.typing import NDArray
from psygnal import Signal

from whitecanvas.backend import Backend
from whitecanvas.layers import _legend
from whitecanvas.layers._base import HoverableDataBoundLayer
from whitecanvas.layers._mixin import EnumArray
from whitecanvas.layers._primitive.line import LineLayerEvents, MultiLine
from whitecanvas.protocols.layer_protocols import MultiLineProtocol
from whitecanvas.types import (
    ArrayLike1D,
    ColorType,
    LineStyle,
    Orientation,
    OrientationLike,
    XYYData,
    _Void,
)
from whitecanvas.utils.normalize import as_any_1d_array, as_array_1d, as_color_array

_void = _Void()


class ErrorbarsEvents(LineLayerEvents):
    capsize = Signal(float)


class Errorbars(MultiLine, HoverableDataBoundLayer[MultiLineProtocol, XYYData]):
    """Errorbars layer (parallel lines with caps)."""

    events: ErrorbarsEvents
    _events_class = ErrorbarsEvents

    def __init__(
        self,
        t: ArrayLike1D,
        edge_low: ArrayLike1D,
        edge_high: ArrayLike1D,
        orient: OrientationLike = Orientation.VERTICAL,
        *,
        name: str | None = None,
        color: ColorType = "black",
        alpha: float = 1,
        width: float = 1,
        style: LineStyle | str = LineStyle.SOLID,
        antialias: bool = True,
        capsize: float = 0.0,
        backend: Backend | str | None = None,
    ):
        t0 = as_array_1d(t)
        y0 = as_array_1d(edge_low)
        y1 = as_array_1d(edge_high)
        if not (t0.size == y0.size == y1.size):
            raise ValueError(
                "Expected all arrays to have the same size, "
                f"got {t0.size}, {y0.size}, {y1.size}"
            )
        if capsize < 0:
            raise ValueError(f"Capsize must be non-negative, got {capsize!r}")
        ori = Orientation.parse(orient)
        if ori is Orientation.VERTICAL:
            data = _xyy_to_segments(t0, y0, y1, capsize)
        else:
            data = _yxx_to_segments(t0, y0, y1, capsize)
        self._orient = ori
        self._capsize = capsize
        self._data = XYYData(t0, y0, y1)
        super().__init__(
            data, name=name, color=color, width=width, style=style,
            antialias=antialias, backend=backend,
        )  # fmt: skip
        self.update(
            color=color, width=width, style=style, alpha=alpha,
            antialias=antialias, capsize=capsize
        )  # fmt: skip

    @property
    def data(self) -> XYYData:
        """Data for this layer."""
        return self._get_layer_data()

    @data.setter
    def data(self, data):
        """Set the data for this layer."""
        self._set_layer_data(self._norm_layer_data(data))
        self.events.data.emit(data)

    @classmethod
    def empty(
        cls,
        orient: OrientationLike = Orientation.VERTICAL,
        name: str | None = None,
        backend: Backend | str | None = None,
    ) -> Errorbars:
        """Return an Errorbars instance with no component."""
        return Errorbars([], [], [], name=name, orient=orient, backend=backend)

    @classmethod
    def empty_v(
        cls,
        name: str | None = None,
        backend: Backend | str | None = None,
    ) -> Errorbars:
        """Return a vertical Errorbars instance with no component."""
        return cls.empty(Orientation.VERTICAL, name=name, backend=backend)

    @classmethod
    def empty_h(
        cls, name: str | None = None, backend: Backend | str | None = None
    ) -> Errorbars:
        """Return a horizontal Errorbars instance with no component."""
        return cls.empty(Orientation.HORIZONTAL, name=name, backend=backend)

    def _get_layer_data(self) -> XYYData:
        """Current data of the layer."""
        return self._data

    def _norm_layer_data(self, data: Any) -> XYYData:
        x0, y0, y1 = self.data
        t, edge_low, edge_high = data
        if t is not None:
            x0 = as_array_1d(t)
        if edge_low is not None:
            y0 = as_array_1d(edge_low)
        if edge_high is not None:
            y1 = as_array_1d(edge_high)
        if x0.size != y0.size or x0.size != y1.size:
            raise ValueError(
                f"Expected data to have the same size, got {x0.size}, {y0.size}"
            )
        return XYYData(x0, y0, y1)

    def _set_layer_data(self, data: XYYData):
        t, y0, y1 = data
        if self._orient.is_vertical:
            segs = _xyy_to_segments(t, y0, y1, self.capsize)
        else:
            segs = _yxx_to_segments(t, y0, y1, self.capsize)
        super()._set_layer_data(segs)
        self._data = data

    def set_data(
        self,
        t: ArrayLike1D | None = None,
        edge_low: ArrayLike1D | None = None,
        edge_high: ArrayLike1D | None = None,
    ):
        self.data = t, edge_low, edge_high

    @property
    def ndata(self) -> int:
        """Number of errorbars."""
        return self.data[0].size

    @property
    def orient(self) -> Orientation:
        """Orientation of the error bars."""
        return self._orient

    @property
    def capsize(self) -> float:
        """Size of the cap of the line edges."""
        return self._capsize

    @capsize.setter
    def capsize(self, capsize: float):
        if capsize < 0:
            raise ValueError(f"Capsize must be non-negative, got {capsize!r}")
        self._capsize = capsize
        with self.events.data.blocked():
            self.set_data(*self._data)
        self.events.capsize.emit(capsize)

    @property
    def color(self) -> NDArray[np.float32]:
        """Colors of the error bars."""
        cols = super().color
        if self._has_caps():
            return cols[: self.ndata]
        return cols

    @color.setter
    def color(self, color: ColorType | Sequence[ColorType]):
        ndata = self.ndata
        col = as_color_array(color, ndata)
        if self._has_caps():
            col = np.concatenate([col] * 3, axis=0)
        MultiLine.color.fset(self, col)

    @property
    def width(self) -> NDArray[np.float32]:
        """Widths of the error bars."""
        widths = super().width
        if self._has_caps():
            return widths[: self.ndata]
        return widths

    @width.setter
    def width(self, width: float | Sequence[float]):
        ndata = self.ndata
        widths = as_any_1d_array(width, ndata, dtype=np.float32)
        if self._has_caps():
            widths = np.concatenate([widths] * 3)
        MultiLine.width.fset(self, widths)

    @property
    def style(self) -> EnumArray[LineStyle]:
        """Styles of the error bars."""
        styles = super().style
        if self._has_caps():
            return np.array(styles[: self.ndata], dtype=object)
        return np.array(styles, dtype=object)

    @style.setter
    def style(self, style: LineStyle | Sequence[LineStyle]):
        ndata = self.ndata
        styles = as_any_1d_array(style, ndata, dtype=object)
        if self._has_caps():
            styles = np.concatenate([styles] * 3)
        MultiLine.style.fset(self, styles)

    def update(
        self,
        *,
        color: ColorType | _Void = _void,
        width: float | _Void = _void,
        style: str | LineStyle | _Void = _void,
        alpha: float | _Void = _void,
        antialias: bool | _Void = _void,
        capsize: float | _Void = _void,
    ):
        super().update(
            color=color, width=width, style=style, alpha=alpha, antialias=antialias
        )
        if capsize is not _void:
            self.capsize = capsize
        return self

    def _has_caps(self) -> bool:
        return self.nlines > self.ndata

    def with_hover_text(self, text: str | Iterable[Any]) -> Errorbars:
        """Add hover text to the errorbars."""
        if isinstance(text, str):
            texts = [text] * self.ndata
        else:
            texts = [str(t) for t in text]
        if self._has_caps():
            texts = texts + texts + texts
        return super().with_hover_text(text)

    def _as_legend_item(self):
        if self.nlines == 0:
            return _legend.EmptyLegendItem()
        return _legend.ErrorLegendItem(
            self.color[0], self.width[0], self.style[0], self.capsize
        )


def _xyy_to_segments(
    x: ArrayLike1D,
    y0: ArrayLike1D,
    y1: ArrayLike1D,
    capsize: float,
):
    """
    ──┬──  <-- y1
      │
      │
    ──┴──  <-- y0
      ↑
      x
    """
    starts = np.stack([x, y0], axis=1)
    ends = np.stack([x, y1], axis=1)
    return _to_segments(starts, ends, capsize)


def _yxx_to_segments(
    y: ArrayLike1D,
    x0: ArrayLike1D,
    x1: ArrayLike1D,
    capsize: float,
):
    starts = np.stack([x0, y], axis=1)
    ends = np.stack([x1, y], axis=1)
    return _to_segments(starts, ends, capsize)


def _to_segments(starts, ends, capsize: float):
    segments = np.stack([starts, ends], axis=1)
    if capsize > 0:
        _c = np.array([capsize / 2, 0])
        cap0 = np.stack([starts - _c, starts + _c], axis=1)
        cap1 = np.stack([ends - _c, ends + _c], axis=1)
    else:
        cap0 = []
        cap1 = []
    return list(segments) + list(cap0) + list(cap1)



from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Generic,
    Iterable,
    TypeVar,
    Union,
    overload,
)

import numpy as np
from cmap import Color, Colormap
from numpy.typing import NDArray

from whitecanvas import layers as _l
from whitecanvas import theme
from whitecanvas.backend import Backend
from whitecanvas.layers import _legend, _mixin
from whitecanvas.layers import group as _lg
from whitecanvas.layers._legend import LegendItem
from whitecanvas.layers.tabular import _jitter, _shared
from whitecanvas.layers.tabular import _plans as _p
from whitecanvas.layers.tabular._df_compat import DataFrameWrapper
from whitecanvas.types import (
    ColormapType,
    ColorType,
    Hatch,
    KdeBandWidthType,
    LineStyle,
    Orientation,
    Symbol,
    _Void,
)
from whitecanvas.utils.type_check import is_real_number

if TYPE_CHECKING:
    from typing_extensions import Self

    from whitecanvas.canvas import CanvasBase

_DF = TypeVar("_DF")
_Cols = Union[str, "tuple[str, ...]"]
_void = _Void()


class _MarkerLikeMixin:
    _source: DataFrameWrapper[_DF]

    @overload
    def update_color(self, value: ColorType) -> Self: ...

    @overload
    def update_color(
        self, by: str | Iterable[str], palette: ColormapType | None = None,
    ) -> Self:  # fmt: skip
        ...

    def update_color(self, by, /, palette=None) -> Self:
        """
        Update colors by a constant value or according to a column.

        >>> layer.update_color("red")  # set all components to red
        >>> layer.update_color("var")  # set color according to the column "var"
        >>> layer.update_color("var", palette=["pink", "#00FF3E"])  # use palette
        """
        cov = _shared.ColumnOrValue(by, self._source)
        if cov.is_column:
            color_by = _p.ColorPlan.from_palette(cov.columns, palette=palette)
        else:
            if palette is not None:
                raise TypeError("`palette` can only be used with a column name.")
            color_by = _p.ColorPlan.from_const(Color(cov.value))
        self._apply_color(color_by.map(self._source))
        self._color_by = color_by
        return self

    @overload
    def update_alpha(self, value: float) -> Self: ...

    @overload
    def update_alpha(
        self,
        by: str | Iterable[str],
        map_from: tuple[float, float] | None = None,
        map_to: tuple[float, float] = (0.2, 1.0),
    ) -> Self: ...

    def update_alpha(self, by, /, map_from=None, map_to=(0.2, 1.0)) -> Self:
        if isinstance(by, str):
            alpha_by = _p.AlphaPlan.from_range(by, range=map_to, domain=map_from)
        else:
            if map_from is not None:
                raise TypeError("`map_from` can only be used with a column name.")
            alpha_by = _p.AlphaPlan.from_const(float(by))
        self._apply_alpha(alpha_by.map(self._source))
        self._alpha_by = alpha_by
        return self

    def update_colormap(
        self,
        by: str,
        cmap: ColormapType = "viridis",
        clim: tuple[float, float] | None = None,
    ) -> Self:
        """
        Update colors according to a column with numeric values using a colormap.

        >>> layer.update_colormap("var", cmap="viridis")  # map "var" column
        >>> layer.update_colormap("var", cmap="viridis", clim=(0, 10))  # set contrast
        """
        if not isinstance(by, str):
            raise ValueError("Can only colormap by a single column.")
        color_by = _p.ColormapPlan.from_colormap(by, cmap=Colormap(cmap), clim=clim)
        self._apply_color(color_by.map(self._source))
        self._color_by = color_by
        return self

    @overload
    def update_width(self, value: float) -> Self: ...

    @overload
    def update_width(
        self, by: str, map_from: tuple[float, float] | None = None,
        map_to: tuple[float, float] = (1.0, 4.0),
    ) -> Self:  # fmt: skip
        ...

    def update_width(self, by, /, map_from=None, map_to=(1.0, 4.0)) -> Self:
        """
        Update the line widths by a constant number or according to a column.

        >>> layer.update_width(2.0)  # set widths of all components to 2.0
        >>> layer.update_width("var")  # set width according to the column "var"

        Parameters
        ----------
        map_from : tuple of float, optional
            Limits of values that will be linearly mapped to the edge width. Data
            points outside this range will be clipped. If not specified, the min/max
            of the data will be used.
        map_to : tuple of float, optional
            Minimum and maximum size of the markers.
        """
        if isinstance(by, str):
            width_by = _p.WidthPlan.from_range(by, range=map_to, domain=map_from)
        else:
            if map_from is not None:
                raise TypeError("`map_from` can only be used with a column name.")
            width_by = _p.WidthPlan.from_const(float(by))
        self._apply_width(width_by.map(self._source))
        self._width_by = width_by
        return self

    @overload
    def update_style(self, value: ColorType) -> Self: ...

    @overload
    def update_style(
        self,
        by: str | Iterable[str],
        palette: ColormapType | None = None,
    ) -> Self: ...

    def update_style(self, by, /, palette=None) -> Self:
        cov = _shared.ColumnOrValue(by, self._source)
        if cov.is_column:
            style_by = _p.StylePlan.new(cov.columns, palette)
        else:
            if palette is not None:
                raise TypeError("`palette` can only be used with a column name.")
            style_by = _p.StylePlan.from_const(LineStyle(cov.value))
        self._apply_style(style_by.map(self._source))
        self._style_by = style_by
        return self

    def _init_color_for_canvas(self, color, alpha, canvas: CanvasBase):
        if color is not None and not self._color_by.is_const():
            self.update_color(self._color_by.by, palette=canvas._color_palette)
        elif color is None:
            self.update_color(canvas._color_palette.next())
        if alpha is not None:
            self.update_alpha(alpha)
        return self

    def _apply_color(self, color):
        """Set color array to the layer."""
        raise NotImplementedError

    def _apply_width(self, width):
        """Set width array to the layer."""
        raise NotImplementedError

    def _apply_alpha(self, alpha):
        """Set alpha array to the layer."""
        raise NotImplementedError

    def _apply_style(self, style):
        """Set style array to the layer."""
        raise NotImplementedError


class DFMarkers(
    _shared.DataFrameLayerWrapper[_lg.MarkerCollection, _DF], _MarkerLikeMixin
):
    def __init__(
        self,
        source: DataFrameWrapper[_DF],
        x: _jitter.JitterBase,
        y: _jitter.JitterBase,
        *,
        color: str | tuple[str, ...] | None = None,
        hatch: str | tuple[str, ...] | None = None,
        symbol: str | tuple[str, ...] | None = None,
        size: str | None = None,
        name: str | None = None,
        backend: str | Backend | None = None,
    ):
        self._x = x
        self._y = y
        self._color_by: _p.ColorPlan | _p.ColormapPlan = _p.ColorPlan.default()
        self._edge_color_by: _p.ColorPlan | _p.ColormapPlan = _p.ColorPlan.default()
        self._hatch_by = _p.HatchPlan.default()
        self._size_by = _p.SizePlan.default()
        self._symbol_by = _p.SymbolPlan.default()
        self._width_by = _p.WidthPlan.default()

        base = _lg.MarkerCollection.from_arrays(
            x.map(source), y.map(source), name=name, backend=backend
        )

        super().__init__(base, source)
        if color is not None:
            self.update_color(color)
        if hatch is not None:
            self.update_hatch(hatch)
        if symbol is not None:
            self.update_symbol(symbol)
        if size is not None:
            self.update_size(size)
        else:
            self.update_size(theme.get_theme().markers.size)

        # set default hover text
        self.with_hover_template(default_template(source.iter_items()))

    def _apply_color(self, color):
        self.base.face.color = color

    def _apply_width(self, width):
        self.base.with_edge(color=_void, width=width, style=_void)

    def _apply_alpha(self, alpha):
        self.base.face.alpha = alpha
        self.base.edge.alpha = alpha

    def _apply_style(self, style):
        self.base.with_edge(color=_void, width=_void, style=style)

    def update_edge_color(self, by: str | Iterable[str], palette=None) -> Self:
        cov = _shared.ColumnOrValue(by, self._source)
        if cov.is_column:
            color_by = _p.ColorPlan.from_palette(cov.columns, palette=palette)
        else:
            color_by = _p.ColorPlan.from_const(Color(cov.value))
        colors = color_by.map(self._source)
        self.base.edge.color = colors
        self._edge_color_by = color_by
        return self

    def update_edge_colormap(
        self,
        by: str,
        cmap: ColormapType | None = None,
        clim: tuple[float, float] | None = None,
    ) -> Self:
        if not isinstance(by, str):
            raise ValueError("Can only colormap by a single column.")
        if cmap is None:
            cmap = Colormap("viridis")
        else:
            cmap = Colormap(cmap)
        color_by = _p.ColormapPlan.from_colormap(by, cmap=cmap, clim=clim)
        colors = color_by.map(self._source)
        self.base.edge.color = colors
        self._edge_color_by = color_by
        return self

    def update_hatch(self, by: str | Iterable[str], palette=None) -> Self:
        cov = _shared.ColumnOrValue(by, self._source)
        if cov.is_column:
            hatch_by = _p.HatchPlan.new(cov.columns, values=palette)
        else:
            hatch_by = _p.HatchPlan.from_const(Hatch(cov.value))
        hatches = hatch_by.map(self._source)
        self.base.face.hatch = hatches
        self._hatch_by = hatch_by
        return self

    @overload
    def update_size(self, value: float) -> Self: ...

    @overload
    def update_size(
        self, by: str, /, map_from: tuple[float, float] | None = None,
        map_to: tuple[float, float] = (3, 15),
    ) -> Self:  # fmt: skip
        ...

    def update_size(self, by, /, map_from=None, map_to=(3, 15)):
        """
        Set the size of the markers.

        >>> layer.update_size(2.0)  # set size of all components to 2.0
        >>> layer.update_size("var")  # set sizes according to the column "var"

        Parameters
        ----------
        map_from : tuple of float, optional
            Limits of values that will be linearly mapped to the marker size. Data
            points outside this range will be clipped. If not specified, the min/max
            of the data will be used.
        map_to : tuple of float, optional
            Minimum and maximum size of the markers.
        """
        if isinstance(by, str):
            size_by = _p.SizePlan.from_range(by, range=map_to, domain=map_from)
        else:
            size_by = _p.SizePlan.from_const(float(by))
        self.base.size = size_by.map(self._source)
        self._size_by = size_by
        return self

    @overload
    def update_symbol(self, value: str | Symbol) -> Self: ...

    @overload
    def update_symbol(
        self, by: str | Iterable[str] | None = None, symbols=None
    ) -> Self: ...

    def update_symbol(self, by, /, symbols=None) -> Self:
        cov = _shared.ColumnOrValue(by, self._source)
        if cov.is_column:
            symbol_by = _p.SymbolPlan.new(cov.columns, values=symbols)
        else:
            symbol_by = _p.SymbolPlan.from_const(Symbol(by))
        self.base.symbol = symbol_by.map(self._source)
        self._symbol_by = symbol_by
        return self

    def with_edge(
        self,
        *,
        color: ColorType | None = None,
        width: float = 1.0,
        style: LineStyle | str = LineStyle.SOLID,
    ) -> Self:
        """Add edge to the markers."""
        if color is not None:
            self.update_edge_color(color)
        self.update_width(width)
        self.base.edge.style = LineStyle(style)
        return self

    def move(self, dx: float = 0.0, dy: float = 0.0, autoscale: bool = True) -> Self:
        """Add a constant shift to the layer."""
        _old_data = self.base.data
        self.base.set_data(xdata=_old_data.x + dx, ydata=_old_data.y + dy)
        if autoscale and (canvas := self._canvas_ref()):
            canvas._autoscale_for_layer(self, pad_rel=0.025)
        return self

    def as_edge_only(
        self,
        width: float = 3.0,
        style: str | LineStyle = LineStyle.SOLID,
    ) -> Self:
        """
        Convert the markers to edge-only mode.

        This method will set the face color to transparent and the edge color to the
        current face color.

        Parameters
        ----------
        width : float, default 3.0
            Width of the edge.
        style : str or LineStyle, default LineStyle.SOLID
            Line style of the edge.
        """
        for layer in self.base.iter_children():
            layer.as_edge_only(width=width, style=style)
        self._edge_color_by = self._color_by
        self._color_by = _p.ColorPlan.from_const("#00000000")
        return self

    def with_hover_template(self, template: str) -> Self:
        """Set the hover tooltip template for the layer."""
        extra = dict(self._source.iter_items())
        self.base.with_hover_template(template, extra=extra)
        return self

    def _as_legend_item(self) -> LegendItem:
        items = []
        color_default = Color("transparent")
        symbol_default = Symbol.CIRCLE
        size_default = 8
        edge_info = self._base_layer.edge._as_legend_info()
        if self._symbol_by.is_const():
            symbol_default = self._symbol_by.get_const_value()
        if self._size_by.is_const():
            size_default = self._size_by.get_const_value()
        if self._color_by.is_const():
            color_default = self._color_by.map(self._source)[0]
        elif isinstance(self._color_by, _p.ColorPlan):
            color_entries = self._color_by.to_entries(self._source)
            items.append((", ".join(self._color_by.by), _legend.TitleItem()))
            for label, color in color_entries:
                item = (
                    label,
                    _legend.MarkersLegendItem(
                        symbol_default, size_default, _legend.FaceInfo(color), edge_info
                    ),
                )
                items.append(item)
        elif isinstance(self._color_by, _p.MapPlan):
            if _map := self._color_by.get_colormap_map():
                items.append((", ".join(self._color_by._on), _legend.TitleItem()))
                for color, value in _map.create_samples(self._source):
                    item = (
                        _safe_str(value),
                        _legend.MarkersLegendItem(
                            symbol_default,
                            size_default,
                            _legend.FaceInfo(color),
                            edge_info,
                        ),
                    )
                    items.append(item)

        if self._hatch_by.is_not_const():
            hatch_entries = self._hatch_by.to_entries(self._source)
            items.append((", ".join(self._hatch_by.by), _legend.TitleItem()))
            for label, hatch in hatch_entries:
                item = (
                    label,
                    _legend.MarkersLegendItem(
                        symbol_default,
                        size_default,
                        _legend.FaceInfo(color_default, hatch),
                        edge_info,
                    ),
                )
                items.append(item)
        if self._symbol_by.is_not_const():
            symbol_entries = self._symbol_by.to_entries(self._source)
            items.append((", ".join(self._symbol_by.by), _legend.TitleItem()))
            for label, symbol in symbol_entries:
                item = (
                    label,
                    _legend.MarkersLegendItem(
                        symbol, size_default, _legend.FaceInfo(color_default), edge_info
                    ),
                )
                items.append(item)
        if self._size_by.is_not_const():
            if _map := self._size_by.get_ranged_map():
                items.append((", ".join(self._size_by._on), _legend.TitleItem()))
                for size, value in _map.create_samples(self._source):
                    item = (
                        _safe_str(value),
                        _legend.MarkersLegendItem(
                            symbol_default,
                            size,
                            _legend.FaceInfo(color_default),
                            edge_info,
                        ),
                    )
                    items.append(item)
        return _legend.LegendItemCollection(items)


class DFMarkerGroups(DFMarkers):
    def __init__(self, *args, orient: Orientation = Orientation.VERTICAL, **kwargs):
        super().__init__(*args, **kwargs)
        self._orient = Orientation.parse(orient)

    @property
    def orient(self) -> Orientation:
        """Orientation of the plot."""
        return self._orient

    def move(self, shift: float = 0.0, autoscale: bool = True) -> Self:
        """Add a constant shift to the layer."""
        _old_data = self._base_layer.data
        if self.orient.is_vertical:
            self._base_layer.set_data(xdata=_old_data.x + shift)
        else:
            self._base_layer.set_data(ydata=_old_data.y + shift)
        if autoscale and (canvas := self._canvas_ref()):
            canvas._autoscale_for_layer(self, pad_rel=0.025)
        return self


class DFBars(
    _shared.DataFrameLayerWrapper[_l.Bars[_mixin.MultiFace, _mixin.MultiEdge], _DF],
    _MarkerLikeMixin,
    Generic[_DF],
):
    def __init__(
        self,
        source: DataFrameWrapper[_DF],
        x,
        y,
        bottom=None,
        color: str | tuple[str, ...] | None = None,
        hatch: str | tuple[str, ...] | None = None,
        stackby: tuple[str, ...] = (),
        name: str | None = None,
        orient: Orientation = Orientation.VERTICAL,
        extent: float = 0.8,
        backend: str | Backend | None = None,
    ):
        splitby = _shared.join_columns(color, hatch, stackby, source=source)
        self._color_by = _p.ColorPlan.default()
        self._hatch_by = _p.HatchPlan.default()
        self._style_by = _p.StylePlan.default()
        self._splitby = splitby
        self._stackby = stackby

        base = _l.Bars(
            x, y, bottom=bottom, name=name, orient=orient, extent=extent,
            backend=backend
        ).with_face_multi().with_edge_multi()  # fmt: skip
        super().__init__(base, source)
        if color is not None:
            self.update_color(color)
        if hatch is not None:
            self.update_hatch(hatch)
        self.with_hover_template(default_template(source.iter_items()))

    @property
    def orient(self) -> Orientation:
        """Orientation of the plot."""
        return self.base.orient

    @classmethod
    def from_table(
        cls,
        df: DataFrameWrapper[_DF],
        x: str | _jitter.JitterBase,
        y: str | _jitter.JitterBase,
        *,
        color: str | tuple[str, ...] | None = None,
        hatch: str | tuple[str, ...] | None = None,
        name: str | None = None,
        extent: float = 0.8,
        orient: Orientation = Orientation.VERTICAL,
        backend: str | Backend | None = None,
    ) -> DFBars[_DF]:
        if isinstance(x, _jitter.JitterBase):
            xj = x
        else:
            xj = _jitter.IdentityJitter(x)
        if isinstance(y, _jitter.JitterBase):
            yj = y
        else:
            yj = _jitter.IdentityJitter(y)
        x0 = xj.map(df)
        y0 = yj.map(df)
        return DFBars(
            df, x0, y0, name=name, color=color, hatch=hatch, extent=extent,
            orient=orient, backend=backend,
        )  # fmt: skip

    @classmethod
    def from_table_stacked(
        cls,
        df: DataFrameWrapper[_DF],
        x: str | _jitter.JitterBase,
        y: str | _jitter.JitterBase,
        stackby: str | tuple[str, ...] | None,
        *,
        color: str | tuple[str, ...] | None = None,
        hatch: str | tuple[str, ...] | None = None,
        name: str | None = None,
        extent: float = 0.8,
        orient: Orientation = Orientation.VERTICAL,
        backend: str | Backend | None = None,
    ) -> DFBars[_DF]:
        if stackby is None:
            stackby = _shared.join_columns(color, hatch, source=df)
        x0, y0, b0 = _shared.resolve_stacking(df, x, y, stackby)
        return DFBars(
            df, x0, y0, b0, color=color, hatch=hatch, stackby=stackby, name=name,
            extent=extent, orient=orient, backend=backend,
        )  # fmt: skip

    def _apply_color(self, color):
        self.base.face.color = color

    def _apply_width(self, width):
        self._base_layer.with_edge_multi(color=_void, width=width, style=_void)

    def _apply_alpha(self, alpha):
        self.base.face.alpha = alpha
        self.base.edge.alpha = alpha

    def _apply_style(self, style):
        self._base_layer.with_edge_multi(color=_void, width=_void, style=style)

    def update_hatch(self, by: str | Iterable[str], choices=None) -> Self:
        cov = _shared.ColumnOrValue(by, self._source)
        if cov.is_column:
            if set(cov.columns) > set(self._splitby):
                raise ValueError(f"Cannot hatch by a column other than {self._splitby}")
            hatch_by = _p.HatchPlan.new(cov.columns, values=choices)
        else:
            hatch_by = _p.HatchPlan.from_const(Hatch(cov.value))
        self._base_layer.face.hatch = hatch_by.map(self._source)
        self._hatch_by = hatch_by
        return self

    def as_edge_only(
        self,
        width: float = 3.0,
        style: str | LineStyle = LineStyle.SOLID,
    ) -> Self:
        """
        Convert the bars to edge-only mode.

        This method will set the face color to transparent and the edge color to the
        current face color.

        Parameters
        ----------
        width : float, default 3.0
            Width of the edge.
        style : str or LineStyle, default LineStyle.SOLID
            Line style of the edge.
        """
        self.base.as_edge_only(width=width, style=style)
        return self

    def with_hover_template(self, template: str) -> Self:
        """Set the hover tooltip template for the layer."""
        extra = dict(self._source.iter_items())
        self.base.with_hover_template(template, extra=extra)
        return self

    def _as_legend_item(self) -> LegendItem:
        items = []
        color_default = theme.get_theme().background_color
        edge_info = self._base_layer.edge._as_legend_info()
        if self._color_by.is_const():
            color_default = self._color_by.map(self._source)[0]
        else:
            color_entries = self._color_by.to_entries(self._source)
            items.append((", ".join(self._color_by.by), _legend.TitleItem()))
            for label, color in color_entries:
                items.append(
                    (label, _legend.BarLegendItem(_legend.FaceInfo(color), edge_info))
                )

        if self._hatch_by.is_not_const():
            hatch_entries = self._hatch_by.to_entries(self._source)
            items.append((", ".join(self._hatch_by.by), _legend.TitleItem()))
            for label, hatch in hatch_entries:
                item = (
                    label,
                    _legend.BarLegendItem(
                        _legend.FaceInfo(color_default, hatch), edge_info
                    ),
                )
                items.append(item)
        return _legend.LegendItemCollection(items)


class DFRug(_shared.DataFrameLayerWrapper[_l.Rug, _DF], _MarkerLikeMixin, Generic[_DF]):
    def __init__(
        self,
        source: DataFrameWrapper[_DF],
        base: _l.Rug,
        color: _Cols | None = None,
        width: str | float | None = None,
        style: str | Iterable[str] | None = None,
    ):
        self._color_by = _p.ColorPlan.default()
        self._width_by = _p.WidthPlan.default()
        self._style_by = _p.StylePlan.default()
        self._scale_by = _p.WidthPlan.default()
        super().__init__(base, source)
        if color is not None:
            self.update_color(color)
        if isinstance(width, str):
            self.update_width(width)
        elif is_real_number(width):
            self.base.width = width
        if style is not None:
            self.update_style(style)
        self.with_hover_template(default_template(source.iter_items()))

    @classmethod
    def from_table(
        cls,
        df: DataFrameWrapper[_DF],
        value: str,
        color: str | None = None,
        width: float = 1.0,
        style: str | None = None,
        low: float = 0.0,
        high: float = 1.0,
        name: str | None = None,
        orient: str | Orientation = Orientation.VERTICAL,
        backend: str | Backend | None = None,
    ) -> DFRug[_DF]:
        ori = Orientation.parse(orient)
        base = _l.Rug(
            df[value], name=name, orient=ori, low=low, high=high, backend=backend,
        )  # fmt: skip
        return cls(df, base, color=color, width=width, style=style)

    @property
    def orient(self) -> Orientation:
        """Orientation of the rugs"""
        return self.base.orient

    def _apply_color(self, color):
        self.base.color = color

    def _apply_alpha(self, alpha):
        self.base.alpha = alpha

    def _apply_width(self, width):
        self.base.width = width

    def _apply_style(self, style):
        self.base.style = style

    # def update_scale(self, by: str | float, align: str = "low") -> Self:
    #     ...

    def update_length(
        self,
        lengths: float | NDArray[np.number],
        *,
        offset: float | None = None,
        align: str = "low",
    ) -> Self:
        """
        Update the length of the rug lines.

        Parameters
        ----------
        lengths : float or array-like
            Length of the rug lines. If a scalar, all the lines have the same length.
            If an array, each line has a different length.
        offset : float, optional
            Offset of the lines. If not given, the mean of the lower and upper bounds is
            used.
        align : {'low', 'high', 'center'}, optional
            How to align the rug lines around the offset. This parameter is defined as
            follows.

            ```
               "low"     "high"    "center"
              ──┴─┴──   ──┬─┬──    ──┼─┼──
            ```
        """
        self.base.update_length(lengths=lengths, offset=offset, align=align)
        return self

    def with_hover_template(self, template: str) -> Self:
        """Set the hover tooltip template for the layer."""
        extra = dict(self._source.iter_items())
        self.base.with_hover_template(template, extra=extra)
        return self


class DFRugGroups(DFRug[_DF]):
    def __init__(
        self,
        source: DataFrameWrapper[_DF],
        base: _l.Rug,
        value: str,
        splitby: tuple[str, ...],
        color: str | tuple[str, ...] | None = None,
        width: str | None = None,
        style: str | tuple[str, ...] | None = None,
        extent: float = 0.8,
    ):
        super().__init__(source, base, color=color, width=width, style=style)
        self._splitby = splitby
        self._value = value
        self._extent = extent

    @property
    def orient(self) -> Orientation:
        """Orientation of the plot."""
        return self.base.orient.transpose()

    def move(self, shift: float = 0.0, autoscale: bool = True) -> Self:
        """Add a constant shift to the layer."""
        _old = self._base_layer.data_full
        self._base_layer.data_full = _old.x, _old.y0 + shift, _old.y1 + shift
        if autoscale and (canvas := self._canvas_ref()):
            canvas._autoscale_for_layer(self, pad_rel=0.025)
        return self

    @classmethod
    def from_table(
        cls,
        df: DataFrameWrapper[_DF],
        jitter: _jitter.CategoricalJitter,
        value: str,
        color: str | None = None,
        width: float = 1.0,
        style: str | None = None,
        extent: float = 0.8,
        name: str | None = None,
        orient: str | Orientation = Orientation.VERTICAL,
        backend: str | Backend | None = None,
    ) -> DFRugGroups[_DF]:
        # orientation of the rugs are always opposite to the orientation of the plot
        ori = Orientation.parse(orient).transpose()
        x = jitter.map(df)
        dx = extent / 2
        base = _l.Rug(
            df[value], orient=ori, low=x - dx, high=x + dx, name=name, backend=backend,
        )  # fmt: skip
        return cls(
            df, base, value, jitter.by, color=color, width=width, style=style,
            extent=extent,
        )  # fmt: skip

    def scale_by_density(
        self,
        *,
        align: str = "center",
        band_width: KdeBandWidthType = "scott",
    ) -> Self:
        """
        Set the height of the lines by density.

        Parameters
        ----------
        align : {'low', 'high', 'center'}, optional
            How to align the rug lines around the offset. This parameter is defined as
            follows.

            ```
               "low"     "high"    "center"
              ──┴─┴──   ──┬─┬──    ──┼─┼──
            ```
        band_width : float, "scott" or "silverman", optional
            Method to calculate the estimator bandwidth.
        """
        from whitecanvas.utils.kde import gaussian_kde

        data_full = self.base.data_full
        densities: list[np.ndarray] = []
        slices: list[np.ndarray] = []
        offsets: list[float] = []
        for _sl, sub in self._source.group_by(self._splitby):
            arr = sub[self._value]
            density = gaussian_kde(arr, band_width)(arr)
            densities.append(density)
            _ar_bool = np.column_stack(
                [self._source[col] == s for col, s in zip(self._splitby, _sl)]
            ).all(axis=1)
            slices.append(_ar_bool)
            offsets.append(data_full.ycenter[_ar_bool].mean())
        density_max = max(d.max() for d in densities)
        normed = [d / density_max * self._extent / 2 for d in densities]

        # sort densities
        normed_sorted = np.empty(self._source.shape[0], dtype=np.float32)
        off_sorted = np.empty_like(normed_sorted)
        for _sl, _norm, _off in zip(slices, normed, offsets):
            normed_sorted[_sl] = _norm
            off_sorted[_sl] = _off

        if align == "low":
            y0 = off_sorted
            y1 = off_sorted + normed_sorted
        elif align == "high":
            y0 = off_sorted - normed_sorted
            y1 = off_sorted
        elif align == "center":
            y0 = off_sorted - normed_sorted
            y1 = off_sorted + normed_sorted
        else:
            raise ValueError(
                f"`align` must be 'low', 'high', or 'center', got {align!r}."
            )
        self.base.data_full = data_full.x, y0, y1
        return self


def default_template(it: Iterable[tuple[str, np.ndarray]], max_rows: int = 10) -> str:
    """
    Default template string for markers

    This template can only be used for those plot that has one tooltip for each data
    point, which includes markers, bars and rugs.
    """
    fmt_list = list[str]()
    for ikey, (key, value) in enumerate(it):
        if not key:
            continue
        if ikey >= max_rows:
            break
        if value.dtype.kind == "f":
            fmt_list.append(f"{key}: {{{key}:.4g}}")
        else:
            fmt_list.append(f"{key}: {{{key}!r}}")
    return "\n".join(fmt_list)


def _safe_str(value) -> str:
    if is_real_number(value):
        return f"{value:.3g}"
    else:
        return str(value)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, SupportsIndex


class DimAxis(ABC):
    def __init__(self, name: str):
        self._name = name

    @property
    def name(self) -> str:
        """Name of the axis."""
        return self._name

    @abstractmethod
    def value(self) -> SupportsIndex:
        """Value of the axis."""

    @abstractmethod
    def set_value(self, value: Any):
        """Set the value of the axis."""


class RangeAxis(DimAxis):
    def __init__(self, name: str, size: int):
        super().__init__(name)
        self._size = size
        self._value = 0

    def value(self) -> SupportsIndex:
        return self._value

    def set_value(self, value: Any):
        v = int(value)
        if not 0 <= v < self._size:
            raise ValueError(
                f"Size of axis {self!r} is {self._size} but got index {value!r}."
            )
        self._value = value

    def size(self) -> int:
        return self._size

    def set_size(self, size: int):
        self._size = size
        self._value = min(self._value, size - 1)


class CategoricalAxis(DimAxis):
    def __init__(self, name: str, categories: list[str]):
        super().__init__(name)
        self._mapper = {c: i for i, c in enumerate(categories)}
        self._value = categories[0]

    def value(self) -> SupportsIndex:
        return self._mapper[self._value]

    def set_value(self, value: Any):
        if value not in self._mapper:
            raise ValueError(
                f"Value must be one of {list(self._mapper)} but got {value!r}."
            )
        self._value = value

    def categories(self) -> list[str]:
        return list(self._mapper.keys())



from __future__ import annotations

import sys
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from IPython import InteractiveShell
    from qtpy.QtWidgets import QApplication
    from wx import App as wxApp


class Application(ABC):
    @abstractmethod
    def get_app(self):
        """Get Application."""

    @abstractmethod
    def run_app(self):
        """Start the event loop."""


class QtApplication(Application):
    _APP: QApplication | None = None

    def get_app(self):
        """Get QApplication."""
        self.gui_qt()
        app = self.instance()
        if app is None:
            app = self.create_application()
        self._APP = app
        return app

    def create_application(self):
        from qtpy.QtCore import Qt
        from qtpy.QtWidgets import QApplication

        QApplication.setAttribute(Qt.ApplicationAttribute.AA_EnableHighDpiScaling)
        return QApplication([])

    def run_app(self):
        """Start the event loop."""
        return self.get_app().exec_()

    def instance(self):
        from qtpy.QtWidgets import QApplication

        return QApplication.instance()

    def gui_qt(self):
        """Call "%gui qt" magic."""
        shell = get_shell()

        if shell and shell.active_eventloop != "qt":
            shell.enable_gui("qt")
        return None


class WxApplication(Application):
    _APP: wxApp | None = None

    def get_app(self):
        """Get WxApplication."""
        self.gui_wx()
        app = self.instance()
        if app is None:
            app = self.create_application()
        self._APP = app
        return app

    def create_application(self):
        try:
            from wx import App as wxApp
        except ImportError:
            from wx import PySimpleApp as wxApp

        app = wxApp()
        app.SetExitOnFrameDelete(True)
        return app

    def run_app(self):
        """Start the event loop."""
        return self.get_app().MainLoop()

    def instance(self):
        import wx

        return wx.GetApp()

    def gui_wx(self):
        """Call "%gui wx" magic."""
        shell = get_shell()

        if shell and shell.active_eventloop != "wx":
            shell.enable_gui("wx")
        return None


class TkApplication(Application):
    _APP: QApplication | None = None

    def get_app(self):
        """Get QApplication."""
        self.gui_tk()
        app = self.instance()
        if app is None:
            app = self.create_application()
        self._APP = app
        return app

    def create_application(self):
        from qtpy.QtCore import Qt
        from qtpy.QtWidgets import QApplication

        QApplication.setAttribute(Qt.ApplicationAttribute.AA_EnableHighDpiScaling)
        return QApplication([])

    def run_app(self):
        """Start the event loop."""
        if self.instance() is None:
            return self.get_app().exec_()

    def instance(self):
        from qtpy.QtWidgets import QApplication

        return QApplication.instance()

    def gui_tk(self):
        """Call "%gui tk" magic."""
        shell = get_shell()

        if shell and shell.active_eventloop != "tk":
            shell.enable_gui("tk")
        return None


class EmptyApplication(Application):
    def get_app(self):
        return None

    def run_app(self):
        return None


def get_app(name: str) -> Application:
    if name == "qt":
        return QtApplication()
    elif name == "wx":
        return WxApplication()
    else:
        return EmptyApplication()


def get_shell() -> InteractiveShell | None:
    """Get ipython shell if available."""
    if "IPython" in sys.modules:
        from IPython import get_ipython

        return get_ipython()
    else:
        return None



from __future__ import annotations

import warnings

import numpy as np
from numpy.typing import NDArray
from plotly import graph_objects as go

from whitecanvas.backend.plotly._base import (
    PlotlyHoverableLayer,
    from_plotly_linestyle,
    to_plotly_linestyle,
)
from whitecanvas.protocols import LineProtocol, MultiLineProtocol, check_protocol
from whitecanvas.types import LineStyle
from whitecanvas.utils.normalize import arr_color, rgba_str_color
from whitecanvas.utils.type_check import is_real_number


@check_protocol(LineProtocol)
class MonoLine(PlotlyHoverableLayer[go.Scatter]):
    def __init__(self, xdata, ydata):
        ndata = len(xdata)
        self._props = {
            "x": xdata,
            "y": ydata,
            "mode": "lines",
            "line": {"color": "blue", "width": 1, "dash": "solid", "simplify": False},
            "type": "scatter",
            "showlegend": False,
            "visible": True,
            "customdata": [""] * ndata,
            "hovertemplate": "%{customdata}<extra></extra>",
        }
        PlotlyHoverableLayer.__init__(self)

    def _plt_get_data(self):
        return self._props["x"], self._props["y"]

    def _plt_set_data(self, xdata, ydata):
        self._props["x"] = xdata
        self._props["y"] = ydata

    def _plt_get_ndata(self) -> int:
        return len(self._props["x"])

    def _plt_get_edge_width(self) -> float:
        return self._props["line"]["width"]

    def _plt_set_edge_width(self, width: float):
        self._props["line"]["width"] = width

    def _plt_get_edge_style(self) -> LineStyle:
        return from_plotly_linestyle(self._props["line"]["dash"])

    def _plt_set_edge_style(self, style: LineStyle):
        self._props["line"]["dash"] = to_plotly_linestyle(style)

    def _plt_get_edge_color(self) -> NDArray[np.float32]:
        return arr_color(self._props["line"]["color"])

    def _plt_set_edge_color(self, color: NDArray[np.float32]):
        self._props["line"]["color"] = rgba_str_color(color)

    def _plt_get_antialias(self) -> bool:
        return not self._props["line"]["simplify"]

    def _plt_set_antialias(self, antialias: bool):
        self._props["line"]["simplify"] = not antialias


@check_protocol(MultiLineProtocol)
class MultiLine(PlotlyHoverableLayer[go.Scatter]):
    def __init__(self, data: list[NDArray[np.floating]]):
        # In plotly, we can break a line into multiple segments by inserting None
        xdata = []
        ydata = []
        for each in data:
            xdata = xdata + each[:, 0].tolist() + [None]
            ydata = ydata + each[:, 1].tolist() + [None]
        # remove last None
        if len(xdata) > 0:
            xdata.pop()
            ydata.pop()
        self._data = data

        self._props = {
            "x": xdata,
            "y": ydata,
            "mode": "lines",
            "line": {
                "color": "blue",
                "width": 1,
                "dash": "solid",
                "simplify": False,
            },
            "type": "scatter",
            "showlegend": False,
            "visible": True,
            "customdata": [""] * len(xdata),
            "hovertemplate": "%{customdata}<extra></extra>",
        }
        x = np.array(xdata, dtype=np.float32)
        self._nan_indices = np.where(np.isnan(x))[0]
        PlotlyHoverableLayer.__init__(self)

    def _plt_get_data(self):
        return self._data

    def _plt_set_data(self, data: list[NDArray[np.floating]]):
        xdata = []
        ydata = []
        for each in data:
            xdata = xdata + each[:, 0].tolist() + [None]
            ydata = ydata + each[:, 1].tolist() + [None]
        # remove last None
        if len(xdata) > 0:
            xdata.pop()
            ydata.pop()
        self._data = data
        self._props["x"] = xdata
        self._props["y"] = ydata

    def _plt_get_ndata(self) -> int:
        return len(self._data)

    def _plt_get_edge_width(self) -> NDArray[np.float32]:
        width = self._props["line"]["width"]
        return np.full(self._plt_get_ndata(), width, dtype=np.float32)

    def _plt_set_edge_width(self, width):
        if is_real_number(width):
            w = width
        else:
            candidates = np.unique(width)
            if len(candidates) == 1:
                w = candidates[0]
            elif len(candidates) == 0:
                w = 0.0
            else:
                _warn_multiple("line widths")
                w = width[0]
        self._props["line"]["width"] = w

    def _plt_get_edge_style(self) -> list[LineStyle]:
        style = from_plotly_linestyle(self._props["line"]["dash"])
        return [style] * self._plt_get_ndata()

    def _plt_set_edge_style(self, style: LineStyle | list[LineStyle]):
        if isinstance(style, LineStyle):
            self._props["line"]["dash"] = to_plotly_linestyle(style)
        else:
            candidates = set(style)
            if len(candidates) == 1:
                self._props["line"]["dash"] = to_plotly_linestyle(style[0])
            elif len(candidates) == 0:
                self._props["line"]["dash"] = "solid"
            else:
                _warn_multiple("line styles")
                self._props["line"]["dash"] = to_plotly_linestyle(style[0])

    def _plt_get_edge_color(self) -> NDArray[np.float32]:
        col = arr_color(self._props["line"]["color"])
        return np.stack([col] * self._plt_get_ndata(), axis=0)

    def _plt_set_edge_color(self, color: NDArray[np.float32]):
        if color.ndim == 1:
            self._props["line"]["color"] = rgba_str_color(color)
        else:
            candidates = np.unique(color, axis=0)
            if len(candidates) == 1:
                self._props["line"]["color"] = rgba_str_color(color[0])
            elif len(candidates) == 0:
                self._props["line"]["color"] = "blue"
            else:
                _warn_multiple("line colors")
                self._props["line"]["color"] = rgba_str_color(color[0])

    def _plt_get_antialias(self) -> bool:
        return not self._props["line"]["simplify"]

    def _plt_set_antialias(self, antialias: bool):
        self._props["line"]["simplify"] = not antialias

    def _update_hover_texts(self, fig):
        if self._hover_texts is None:
            return

        dif = np.diff(self._nan_indices, prepend=0, append=len(self._props["x"]))
        # if x = [1, 2, None, 3, 4, 2, None, 3, 2]
        # then dif = [2, 4, 3]
        dif[1:] -= 1
        if len(self._hover_texts) != len(dif):
            warnings.warn(
                "The length of the hover text does not match the number of lines. "
                "Ignoring.",
                UserWarning,
                stacklevel=2,
            )
            return
        customdata = []
        for t, d in zip(self._hover_texts, dif):
            customdata.extend([t] * d)
            customdata.append("")
        customdata.pop()

        fig.update_traces(
            customdata=customdata,
            selector={"uid": self._props["uid"]},
        )

    def _plt_connect_pick_event(self, callback):
        def _new_cb(indices: list[int]):
            return callback([np.where(self._nan_indices < i)[0].size for i in indices])

        return super()._plt_connect_pick_event(_new_cb)


def _warn_multiple(prop: str):
    return warnings.warn(
        f"plotly does not support multiple {prop}. Set to the first one.",
        UserWarning,
        stacklevel=4,
    )



# modified from:
# https://github.com/pyapp-kit/magicgui/blob/main/docs/scripts/_gen_screenshots.py
import re
import warnings
from pathlib import Path
from textwrap import dedent
from typing import Any
from imageio import imwrite
import matplotlib.pyplot as plt
import mkdocs_gen_files

from whitecanvas.canvas import CanvasGrid, SingleCanvas
from whitecanvas.theme import update_default

DOCS: Path = Path(__file__).parent.parent
CODE_BLOCK = re.compile("``` ?python.*?\n([^`]*)```")

def _exec_code(src: str, ns: dict, dest: str) -> dict[str, Any]:
    try:
        exec(src, ns, ns)
    except NameError as e:
        raise NameError(
            f"Error evaluating code for {dest!r} with namespace {set(ns)!r}:\n\n{src}. "
            f"{e}"
        ) from e
    except Exception as e:
        raise RuntimeError(
            f"Error evaluating code\n\n{src}\n\nfor {dest!r}"
        ) from e
    return ns

def _write_image(src: str, ns: dict, dest: str) -> None:
    ns = _exec_code(src, ns, dest)
    if "grid" in src and isinstance(ns.get("grid", None), CanvasGrid):
        canvas = ns["grid"]
    elif "canvas" in src:
        canvas = ns["canvas"]
    assert isinstance(canvas, (CanvasGrid, SingleCanvas)), type(canvas)
    with mkdocs_gen_files.open(dest, "wb") as f:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            plt.tight_layout()
        img = canvas.screenshot()
        imwrite(f.name, img, format="png")

def _write_html(src: str, ns: dict, dest: str) -> None:
    ns = _exec_code(src, ns, dest)
    if isinstance(ns.get("grid", None), CanvasGrid):
        canvas = ns["grid"]
    else:
        canvas = ns["canvas"]
    assert isinstance(canvas, (CanvasGrid, SingleCanvas)), type(canvas)
    with mkdocs_gen_files.open(dest, "w") as f:
        canvas.to_html(f.name)

EXCLUDE: set[str] = set()


def main() -> None:
    with update_default() as theme:
        theme.canvas_size = (800, 560)
        theme.font.size = 18
    plt.switch_backend("Agg")
    names_found = set[str]()
    for mdfile in sorted(DOCS.rglob("*.md"), reverse=True):
        if mdfile.name in EXCLUDE:
            continue

        md = mdfile.read_text()
        code_blocks = list(CODE_BLOCK.finditer(md))
        namespace = {}
        for match in code_blocks:
            code = dedent(match.group(1)).strip()
            if code.startswith("#!skip"):
                continue
            elif code.startswith("#!name:"):
                if code.endswith(("canvas.show()", "grid.show()")):
                    code = code[:-7]
                line = code.split("\n", 1)[0]
                assert line.startswith("#!name:")
                name = line.split(":", 1)[1].strip()
                if name in names_found:
                    raise ValueError(f"Duplicate name {name!r} in {mdfile}")
                dest = f"_images/{name}.png"
                _write_image(code, namespace, dest)
                names_found.add(f"{name}.png")
            elif code.startswith("#!html:"):
                if code.endswith(("canvas.show()", "grid.show()")):
                    code = code[:-7]
                line = code.split("\n", 1)[0]
                assert line.startswith("#!html:")
                name = line.split(":", 1)[1].strip()
                if name in names_found:
                    raise ValueError(f"Duplicate name {name!r} in {mdfile}")
                dest = f"_images/{name}.html"
                _write_html(code, namespace, dest)
                names_found.add(f"{name}.html")
            else:
                if code.endswith(("canvas.show()", "grid.show()")):
                    code = code[:-7]
                try:
                    exec(code, namespace, namespace)
                except Exception as e:
                    raise RuntimeError(
                        f"Error evaluating code\n\n{code}\n\nfor {dest!r}"
                    ) from e
            # close all if there's more than 10 figures
            if len(plt.get_fignums()) > 10:
                plt.close("all")

main()



import numpy as np

import whitecanvas.plot as plt


def test_functions():
    arr = np.arange(10)
    plt.figure(size=(400, 300))
    plt.line(arr)
    plt.markers(arr)
    plt.bars(arr, arr / 2)
    plt.band(arr, arr / 2, arr * 2)
    plt.errorbars(arr, arr / 2, arr * 2)
    plt.hist(arr)
    plt.spans([[4, 10], [5, 14]])
    plt.infcurve(lambda x: x ** 2)
    plt.infline((0, 0), 40)
    plt.legend()
    plt.xlim()
    plt.xlim(0, 10)
    plt.ylim()
    plt.ylim(0, 10)
    plt.title("Title")
    plt.xlabel("X")
    plt.ylabel("Y")
    plt.xticks()
    plt.yticks()
    plt.show()



from whitecanvas.layers.tabular._box_like import (
    DFBarPlot,
    DFBoxPlot,
    DFPointPlot,
    DFViolinPlot,
)
from whitecanvas.layers.tabular._dataframe import (
    DFHeatmap,
    DFHistograms,
    DFKde,
    DFLines,
    DFMultiHeatmap,
    DFPointPlot2D,
)
from whitecanvas.layers.tabular._df_compat import parse
from whitecanvas.layers.tabular._marker_like import (
    DFBars,
    DFMarkerGroups,
    DFMarkers,
    DFRug,
    DFRugGroups,
)
from whitecanvas.layers.tabular._stackable import DFArea

__all__ = [
    "DFArea",
    "DFBarPlot",
    "DFLines",
    "DFViolinPlot",
    "DFMarkerGroups",
    "DFPointPlot",
    "DFMarkers",
    "DFMultiHeatmap",
    "DFBars",
    "DFBoxPlot",
    "DFHeatmap",
    "DFHistograms",
    "DFKde",
    "DFRug",
    "DFRugGroups",
    "DFPointPlot2D",
    "parse",
]



from whitecanvas.backend.plotly.components3d.canvas import Canvas3D
from whitecanvas.backend.plotly.components3d.line import MonoLine3D
from whitecanvas.backend.plotly.components3d.markers import Markers3D
from whitecanvas.backend.plotly.components3d.mesh import Mesh3D



# This example shows how to display an image when the corresponding marker is picked.

from __future__ import annotations

import numpy as np
from whitecanvas import new_row

def make_images() -> np.ndarray:
    # prepare sample image data
    images = []
    def sig(x, a):
        return np.exp(x/a)/(1 + np.exp(x/a))

    mean_intensities = sig(np.arange(20) - 10, 2)
    xx, yy = np.meshgrid(np.linspace(-1, 1, 30), np.linspace(-1, 1, 30))
    weight = np.exp(-(xx**2 + yy**2)*2)
    for i0 in mean_intensities:
        img = weight * i0 + np.random.normal(scale=0.3, size=(30, 30))
        images.append(img * weight)
    return np.stack(images, axis=0)

def main():
    images = make_images()
    means = np.mean(images, axis=(1, 2))  # calculate mean intensity to plot

    g = new_row(2, backend="matplotlib:qt")

    # markers to be picked
    markers = (
        g.add_canvas(0)
        .update_labels(x="time", y="intensity")
        .add_markers(means, color="black")
        .with_hover_text([f"{m:.3f}" for m in means])
    )

    # image to be displayed
    img_layer = (
        g.add_canvas(1)
        .update_labels(title="image slice")
        .add_image(images[0], cmap="inferno")
    )

    # connect pick event
    @markers.events.clicked.connect
    def _on_pick(indices):
        if len(indices) == 1:
            i = indices[0]
            img_layer.data = images[i]
            img_layer.clim = images[i].min(), images[i].max()

    g.show(block=True)

if __name__ == "__main__":
    main()
